/*
 * $Header: /home/cvs/jakarta-tomcat-jasper/jasper2/src/share/org/apache/jasper/runtime/PageContextImpl.java,v 1.26 2002/10/04 19:21:44 luehe Exp $
 * $Revision: 1.26 $
 * $Date: 2002/10/04 19:21:44 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */

package org.apache.jasper.runtime;

import java.io.*;

import java.util.EmptyStackException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Stack;

import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.ServletException;

import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;

import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.JspFactory;
import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.tagext.BodyContent;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.el.ELException;
import javax.servlet.jsp.el.ExpressionEvaluator;
import javax.servlet.jsp.el.VariableResolver;

import org.apache.jasper.Constants;
import org.apache.jasper.logging.Logger;
import org.apache.jasper.runtime.el.jstl.JSTLVariableResolver;

/**
 * Implementation of the PageContext class from the JSP spec.
 * Also doubles as a VariableResolver for the EL.
 *
 * @author Anil K. Vijendran
 * @author Larry Cable
 * @author Hans Bergsten
 * @author Pierre Delisle
 * @author Mark Roth
 */
public class PageContextImpl 
    extends PageContext 
    implements VariableResolver
{

    Logger.Helper loghelper = new Logger.Helper("JASPER_LOG", "PageContextImpl");
    
    /**
     * The expression evaluator, for evaluating EL expressions.
     */
    private ExpressionEvaluatorImpl expressionEvaluator = null;
    
    /**
     * The variable resolver, for evaluating EL expressions.
     */
    private static JSTLVariableResolver variableResolver =
        new JSTLVariableResolver();
    
    /**
     * Expression evaluator for proprietary EL evaluation.
     * XXX - This should be going away once the EL evaluator moves from
     * the JSTL implementation to its own project.
     */
    private static org.apache.jasper.runtime.el.jstl.Evaluator
        proprietaryEvaluator = new 
        org.apache.jasper.runtime.el.jstl.Evaluator();

    PageContextImpl(JspFactory factory) {
        this.factory = factory;
    }

    public void initialize(Servlet servlet, ServletRequest request,
                           ServletResponse response, String errorPageURL,
                           boolean needsSession, int bufferSize,
                           boolean autoFlush)
        throws IOException, IllegalStateException, IllegalArgumentException
    {
	_initialize(servlet, request, response, errorPageURL, needsSession, bufferSize, autoFlush);
    }

    void _initialize(Servlet servlet, ServletRequest request,
                           ServletResponse response, String errorPageURL,
                           boolean needsSession, int bufferSize,
                           boolean autoFlush)
        throws IOException, IllegalStateException, IllegalArgumentException
    {

	// initialize state

	this.servlet      = servlet;
	this.config	  = servlet.getServletConfig();
	this.context	  = config.getServletContext();
	this.needsSession = needsSession;
	this.errorPageURL = errorPageURL;
	this.bufferSize   = bufferSize;
	this.autoFlush    = autoFlush;
	this.request      = request;
	this.response     = response;

	// setup session (if required)
	if (request instanceof HttpServletRequest && needsSession)
	    this.session = ((HttpServletRequest)request).getSession();

	if (needsSession && session == null)
	    throw new IllegalStateException
                ("Page needs a session and none is available");

        // initialize the initial out ...
        depth = -1;
        if (this.baseOut == null) {
            this.baseOut = _createOut(bufferSize, autoFlush);
        } else {
            this.baseOut.init(response, bufferSize, autoFlush);
        }
        this.out = baseOut;

	if (this.out == null)
	    throw new IllegalStateException("failed initialize JspWriter");

	// register names/values as per spec

	setAttribute(OUT,         this.out);
	setAttribute(REQUEST,     request);
	setAttribute(RESPONSE,    response);

	if (session != null)
	    setAttribute(SESSION, session);

	setAttribute(PAGE,        servlet);
	setAttribute(CONFIG,      config);
	setAttribute(PAGECONTEXT, this);
	setAttribute(APPLICATION,  context);
	
	isIncluded = request.getAttribute(
	    "javax.servlet.include.servlet_path") != null;	    
    }

    public void release() {
        out = baseOut;
	try {
	    if (isIncluded) {
		((JspWriterImpl)out).flushBuffer();
			// push it into the including jspWriter
	    } else {
	        out.flush();
	    }
	} catch (IOException ex) {
	    loghelper.log("Internal error flushing the buffer in release()");
	}
	servlet      = null;
	config	     = null;
	context	     = null;
	needsSession = false;
	errorPageURL = null;
	bufferSize   = JspWriter.DEFAULT_BUFFER;
	autoFlush    = true;
	request      = null;
	response     = null;
        depth = -1;
	baseOut.recycle();
	session      = null;

	attributes.clear();
    }

    public Object getAttribute(String name) {

	if (name == null) throw new NullPointerException("Null name");
	return attributes.get(name);
    }


    public Object getAttribute(String name, int scope) {

	if (name == null) throw new NullPointerException("Null name");

	switch (scope) {
	    case PAGE_SCOPE:
		return attributes.get(name);

	    case REQUEST_SCOPE:
		return request.getAttribute(name);

	    case SESSION_SCOPE:
		if (session == null)
		    throw new IllegalArgumentException(
                        "can't access SESSION_SCOPE without an HttpSession");
		else
		    return session.getAttribute(name);

	    case APPLICATION_SCOPE:
		return context.getAttribute(name);

	    default:
		throw new IllegalArgumentException("Invalid scope");
	}
    }


    public void setAttribute(String name, Object attribute) {

	if (name == null || attribute == null)
	    throw new NullPointerException("Null name or attribute value");

	attributes.put(name, attribute);
    }


    public void setAttribute(String name, Object o, int scope) {

	if (name == null || o == null)
	    throw new NullPointerException("Null name or attribute value");

	switch (scope) {
	    case PAGE_SCOPE:
		attributes.put(name, o);
		break;

	    case REQUEST_SCOPE:
		request.setAttribute(name, o);
		break;

	    case SESSION_SCOPE:
		if (session == null)
		    throw new IllegalArgumentException(
                        "can't access SESSION_SCOPE without an HttpSession");
		else
		    session.setAttribute(name, o);
		break;

	    case APPLICATION_SCOPE:
		context.setAttribute(name, o);
		break;

	    default:
		throw new IllegalArgumentException("Invalid scope");
	}
    }

    public void removeAttribute(String name, int scope) {
	switch (scope) {
	    case PAGE_SCOPE:
		attributes.remove(name);
		break;

	    case REQUEST_SCOPE:
		request.removeAttribute(name);
		break;

	    case SESSION_SCOPE:
		if (session == null)
		    throw new IllegalArgumentException(
                        "can't access SESSION_SCOPE without an HttpSession");
		else
                    session.removeAttribute(name);
                // was:
                //		    session.removeValue(name);
                // REVISIT Verify this is correct - akv
		break;

	    case APPLICATION_SCOPE:
		context.removeAttribute(name);
		break;

	    default:
		throw new IllegalArgumentException("Invalid scope");
	}
    }

    public int getAttributesScope(String name) {
	if (attributes.get(name) != null) return PAGE_SCOPE;

	if (request.getAttribute(name) != null)
	    return REQUEST_SCOPE;

	if (session != null) {
	    if (session.getAttribute(name) != null)
	        return SESSION_SCOPE;
	}

	if (context.getAttribute(name) != null) return APPLICATION_SCOPE;

	return 0;
    }

    public Object findAttribute(String name) {
        Object o = attributes.get(name);
        if (o != null)
            return o;

        o = request.getAttribute(name);
        if (o != null)
            return o;

        if (session != null) {
            o = session.getAttribute(name);
            if (o != null)
                return o;
        }

        return context.getAttribute(name);
    }


    public Enumeration getAttributeNamesInScope(int scope) {
	switch (scope) {
	    case PAGE_SCOPE:
		return attributes.keys();

	    case REQUEST_SCOPE:
		return request.getAttributeNames();

	    case SESSION_SCOPE:
		if (session != null) {
		    return session.getAttributeNames();
		} else
		    throw new IllegalArgumentException(
                        "can't access SESSION_SCOPE without an HttpSession");

	    case APPLICATION_SCOPE:
		return context.getAttributeNames();

	    default:
		throw new IllegalArgumentException("Invalid scope");
	}
    }

    public void removeAttribute(String name) {
	try {
	    removeAttribute(name, PAGE_SCOPE);
	    removeAttribute(name, REQUEST_SCOPE);
	    if( session != null ) {
                removeAttribute(name, SESSION_SCOPE);
            }
            removeAttribute(name, APPLICATION_SCOPE);
        } catch (Exception ex) {
	    // we remove as much as we can, and
	    // simply ignore possible exceptions
	}
    }

    public JspWriter getOut() {
	return out;
    }

    public HttpSession getSession() { return session; }
    public Servlet getServlet() { return servlet; }
    public ServletConfig getServletConfig() { return config; }
    public ServletContext getServletContext() {
	return config.getServletContext();
    }
    public ServletRequest getRequest() { return request; }
    public ServletResponse getResponse() { return response; }
    public Exception getException() { return (Exception)request.getAttribute(EXCEPTION); }
    public Object getPage() { return servlet; }


    private final String getAbsolutePathRelativeToContext(String relativeUrlPath) {
        String path = relativeUrlPath;

        if (!path.startsWith("/")) {
	    String uri = (String) request.getAttribute("javax.servlet.include.servlet_path");
	    if (uri == null)
		uri = ((HttpServletRequest) request).getServletPath();
            String baseURI = uri.substring(0, uri.lastIndexOf('/'));
            path = baseURI+'/'+path;
        }

        return path;
    }

    public void include(String relativeUrlPath)
	    throws ServletException, IOException {

        JspRuntimeLibrary.include((HttpServletRequest) request,
                                  (HttpServletResponse) response,
                                  relativeUrlPath,
				  out,
				  true);
    }

    public void include(String relativeUrlPath, boolean flush) 
	    throws ServletException, IOException {

        JspRuntimeLibrary.include((HttpServletRequest) request,
                                  (HttpServletResponse) response,
                                  relativeUrlPath,
				  out,
				  flush);
    }

    public VariableResolver getVariableResolver() {
	return this;
    }

    public void forward(String relativeUrlPath)
        throws ServletException, IOException
    {
	// JSP.4.5 If the buffer was flushed, throw IllegalStateException
	try {
	    out.clear();
	} catch (IOException ex) {
	    throw new IllegalStateException(Constants.getString(
			"jsp.error.attempt_to_clear_flushed_buffer"));
	}

	// Make sure that the response object is not the wrapper for include
	while (response instanceof ServletResponseWrapperInclude) {
	    response = ((ServletResponseWrapperInclude)response).getResponse();
        }

        String path = getAbsolutePathRelativeToContext(relativeUrlPath);
        String includeUri 
            = (String) request.getAttribute(Constants.INC_SERVLET_PATH);
        if (includeUri != null)
            request.removeAttribute(Constants.INC_SERVLET_PATH);
        try {
            context.getRequestDispatcher(path).forward(request, response);
        } finally {
            if (includeUri != null)
                request.setAttribute(Constants.INC_SERVLET_PATH, includeUri);
	    request.setAttribute(Constants.FORWARD_SEEN, "true");
        }
    }

    protected BodyContentImpl[] outs = new BodyContentImpl[0];
    protected int depth = -1;

    public BodyContent pushBody() {
	return (BodyContent) pushBody(null);
    }

    public JspWriter pushBody(Writer writer) {
        depth++;
        if (depth >= outs.length) {
            BodyContentImpl[] newOuts = new BodyContentImpl[depth + 1];
            for (int i=0; i<outs.length; i++) {
                newOuts[i] = outs[i];
            }
            newOuts[depth] = new BodyContentImpl(out);
            outs = newOuts;
        }

	outs[depth].setWriter(writer);
        out = outs[depth];

        return outs[depth];
    }

    public JspWriter popBody() {
        depth--;
        if (depth >= 0) {
            out = outs[depth];
        } else {
            out = baseOut;
        }
        return out;
    }

    /**
     * Provides programmatic access to the ExpressionEvaluator.
     * The JSP Container must return a valid instance of an
     * ExpressionEvaluator that can parse EL expressions.
     */
    public ExpressionEvaluator getExpressionEvaluator() {
        if( this.expressionEvaluator == null ) {
            this.expressionEvaluator = new ExpressionEvaluatorImpl( this );
            // no need to synchronize - not a big deal even if we create 
            // two of these.
        }
        
        return this.expressionEvaluator;
    }

    public void handlePageException(Exception ex)
        throws IOException, ServletException 
    {
	// Should never be called since handleException() called with a
	// Throwable in the generated servlet.
	handlePageException((Throwable) ex);
    }

    public void handlePageException(Throwable t)
        throws IOException, ServletException 
    {
	if (t == null) throw new NullPointerException("null Throwable");

	if (errorPageURL != null && !errorPageURL.equals("")) {

	    // Set request attributes.
	    // Do not set the javax.servlet.error.exception attribute here
	    // (instead, set in the generated servlet code for the error page)
	    // in order to prevent the ErrorReportValve, which is invoked as
	    // part of forwarding the request to the error page, from 
	    // throwing it if the response has not been committed (the response
	    // will have been committed if the error page is a JSP page).
	    request.setAttribute("javax.servlet.jsp.jspException", t);
	    request.setAttribute("javax.servlet.error.status_code",
	        new Integer(HttpServletResponse.SC_INTERNAL_SERVER_ERROR));
	    request.setAttribute("javax.servlet.error.request_uri",
		((HttpServletRequest) request).getRequestURI());
	    request.setAttribute("javax.servlet.error.servlet_name",
				 config.getServletName());
            try {
                forward(errorPageURL);
            } catch (IllegalStateException ise) {
                include(errorPageURL);
            }

	} else {
            // Otherwise throw the exception wrapped inside a ServletException.
	    // Set the exception as the root cause in the ServletException
	    // to get a stack trace for the real problem
	    if (t instanceof IOException) throw (IOException)t;
	    if (t instanceof ServletException) throw (ServletException)t;
            if (t instanceof RuntimeException) throw (RuntimeException)t;
            if (t instanceof JspException) {
                Throwable rootCause = ((JspException)t).getRootCause();
                if (rootCause != null) {
                    throw new ServletException(t.getMessage(), rootCause);
                } else {
                    throw new ServletException(t);
                }
            }
	    throw new ServletException(t);
	}
    }

    /**
     * VariableResolver interface
     */
    public Object resolveVariable( String pName, Object pContext ) 
        throws ELException
    {
        // Note: pContext will be going away.
        try {
            return PageContextImpl.variableResolver.resolveVariable( 
                pName, this );
        }
        catch( org.apache.jasper.runtime.el.jstl.ELException e ) {
            throw new ELException( e );
        }
    }
    
    /**
     * Proprietary method to evaluate EL expressions.
     * XXX - This method should go away once the EL interpreter moves
     * out of JSTL and into its own project.  For now, this is necessary
     * because the standard machinery is too slow.
     *
     * @param expression The expression to be evaluated
     * @param expectedType The expected resulting type
     * @param pageContext The page context
     * @param functionMap Maps prefix and name to Method
     * @param defaultPrefix Default prefix for this evaluation
     * @return The result of the evaluation
     */
    public static Object proprietaryEvaluate( String expression, 
        Class expectedType, PageContext pageContext, Map functionMap, 
        String defaultPrefix )
        throws ELException
    {
        try {
            return PageContextImpl.proprietaryEvaluator.evaluate( "<unknown>", 
                expression, expectedType, null, pageContext, functionMap, 
                defaultPrefix );
        }
        catch( JspException e ) {
            throw new ELException( e );
        }
    }

    protected JspWriterImpl _createOut(int bufferSize, boolean autoFlush)
        throws IOException, IllegalArgumentException {
        try {
            return new JspWriterImpl(response, bufferSize, autoFlush);
        } catch( Throwable t ) {
            loghelper.log("creating out", t);
            return null;
        }
    }
    
    /*
     * fields
     */

    // per Servlet state

    protected 	        Servlet         servlet;
    protected 		ServletConfig   config;
    protected 		ServletContext  context;

    protected 		JspFactory	factory;

    protected		boolean		needsSession;

    protected		String		errorPageURL;

    protected		boolean		autoFlush;
    protected		int		bufferSize;

    // page scope attributes

    protected transient Hashtable	attributes = new Hashtable(16);

    // per request state

    protected transient ServletRequest	request;
    protected transient ServletResponse response;
    protected transient Object          page;

    protected transient HttpSession	session;

    protected boolean isIncluded;

    // initial output stream

    protected transient JspWriter       out;
    protected transient JspWriterImpl   baseOut;

}
